#include "bpop_main.h"


GameTime gameTime = {
    .fixedDeltaTime = 1.0f / 60.0f,
};

Color bubbleTypeColors[] = {
    COLOR_FROM_HEX(0xaaccff),
    COLOR_FROM_HEX(0xffaa00),
    COLOR_FROM_HEX(0x00ffaa),
    COLOR_FROM_HEX(0xff00aa),
    COLOR_FROM_HEX(0xaa00ff),
};

Model bubbleModel;

float BubbleLevelRadius(int level)
{
    return powf((level + 1) * 20, .75f);
}

void PlayfieldFixedUpdate(Playfield *playfield)
{
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        bubble->sameLevelContact = 0;
    }

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        float r = bubble->radius;

        for (int j = i + 1; j < MAX_BUBBLES; j++)
        {
            Bubble *other = &playfield->bubbles[j];
            if (!other->flagIsActive) continue;
            float otherR = other->radius;
            float sumR2 = (r + otherR) * (r + otherR);
            float d2 = Vector2DistanceSqr(bubble->position, other->position);
            int canMerge = bubble->bubbleLevel == other->bubbleLevel && bubble->bubbleType == other->bubbleType;
            if (d2 < sumR2 * 1.05f)
            {
                if (canMerge)
                {
                    bubble->sameLevelContact = 1;
                    other->sameLevelContact = 1;
                }
                if (canMerge && bubble->bubbleMergeCooldown <= 0.0f 
                    && other->bubbleMergeCooldown <= 0.0f)
                {
                    // merge bubbles
                    bubble->bubbleLevel++;
                    bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
                    bubble->bubbleMergeCooldown = 1.0f;
                    other->flagIsActive = 0;
                }
            }

            if (d2 < sumR2)
            {
                float overlap = r + otherR - sqrtf(d2);
                Vector2 normal = Vector2Normalize(Vector2Subtract(other->position, bubble->position));
                // resolve overlap by moving the bubbles apart
                const float errorCorrection = 0.25f;
                bubble->position = Vector2Subtract(bubble->position, Vector2Scale(normal, overlap * errorCorrection));
                other->position = Vector2Add(other->position, Vector2Scale(normal, overlap * errorCorrection));

                // bounce off each other
                Vector2 relativeVelocity = Vector2Subtract(bubble->velocity, other->velocity);
                float dot = Vector2DotProduct(relativeVelocity, normal);
                if (dot > 0.0f)
                {
                    // DrawLineV(bubble->position, other->position, COLOR_FROM_HEX(0xff0000));
                    float impulse = -dot * 0.85f;
                    bubble->velocity = Vector2Add(bubble->velocity, Vector2Scale(normal, impulse));
                    other->velocity = Vector2Subtract(other->velocity, Vector2Scale(normal, impulse));
                }
            }
        }

        if (!bubble->sameLevelContact)
        {
            bubble->bubbleMergeCooldown = 1.0f;
        }
        else
        {
            bubble->bubbleMergeCooldown -= gameTime.fixedDeltaTime;
        }

        float buoyancy = -20.0f;
        if (bubble->position.y < playfield->waterLevel)
        {
            buoyancy = (playfield->waterLevel - bubble->position.y) * 0.5f;
        }
        bubble->velocity = Vector2Add(bubble->velocity, (Vector2){0, buoyancy});
        bubble->velocity = Vector2Scale(bubble->velocity, 0.92f);
        bubble->position.x += bubble->velocity.x * gameTime.fixedDeltaTime;
        bubble->position.y += bubble->velocity.y * gameTime.fixedDeltaTime;
        if ((bubble->position.x < r && bubble->velocity.x < 0.0f) || 
            (bubble->position.x > playfield->fieldSize.x - r && bubble->velocity.x > 0.0f))
        {
            bubble->velocity.x *= -0.9f;
        }

        bubble->position.x = (bubble->position.x < r) ? r : (bubble->position.x > playfield->fieldSize.x - r) ? playfield->fieldSize.x - r : bubble->position.x;
        // bubble->position.y = (bubble->position.y < r) ? r : (bubble->position.y > playfield->fieldSize.y - r) ? playfield->fieldSize.y - r : bubble->position.y;

        // debug velocity
        // DrawLineV(bubble->position, Vector2Add(bubble->position, Vector2Scale(bubble->velocity, 1.0f)), COLOR_FROM_HEX(0xff0000));
    }
}

void PlayfieldTryAddBubble(Playfield *playfield, Vector2 position)
{
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive)
        {
            bubble->flagIsActive = 1;
            bubble->position = position;
            bubble->velocity = (Vector2){GetRandomValue(-100, 100), GetRandomValue(-100, 100)};
            bubble->bubbleType = playfield->nextBubbleType;
            bubble->bubbleLevel = playfield->nextBubbleLevel;
            bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
            break;
        }
    }

    playfield->nextBubbleType = GetRandomValue(0, MAX_BUBBLE_TYPES - 1);
    playfield->nextBubbleLevel = GetRandomValue(0, 3);
}

Vector2 PlayfieldPositionToSpawnPosition(Playfield *playfield, Vector2 position)
{
    Vector2 spawnPosition = position;
    spawnPosition.y = BubbleLevelRadius(5);
    return spawnPosition;
}
Vector2 PlayfieldScreenToSpawnPosition(Playfield *playfield, Camera3D camera, Vector2 screenPosition)
{
    Vector3 cursorPosition = GetScreenToWorldRay(screenPosition, camera).position;
    cursorPosition.x += playfield->fieldSize.x / 2;
    cursorPosition.y += playfield->fieldSize.y / 2;

    Vector2 pos = {cursorPosition.x, cursorPosition.y};
    return PlayfieldPositionToSpawnPosition(playfield, pos);
}

void PlayfieldDraw(Playfield *playfield, Camera3D camera)
{
    float bubbleExtraRadius = 5.0f;
    Vector2 mousePos = GetMousePosition();
    Vector2 spawnPosition = PlayfieldScreenToSpawnPosition(playfield, camera, mousePos);
    DrawCube((Vector3){0, 0, 0}, playfield->fieldSize.x, playfield->fieldSize.y, 0, COLOR_FROM_HEX(0xaabbee));
    DrawCube((Vector3){0, (playfield->waterLevel - playfield->fieldSize.y) * 0.5f, 0}, playfield->fieldSize.x, 
        playfield->waterLevel, 0, COLOR_FROM_HEX(0x225588));

    DrawModel(bubbleModel, 
        (Vector3){spawnPosition.x - playfield->fieldSize.x * 0.5f, spawnPosition.y - playfield->fieldSize.y * 0.5f, 0},
        BubbleLevelRadius(playfield->nextBubbleLevel) + bubbleExtraRadius, bubbleTypeColors[playfield->nextBubbleType]);
    // DrawRectangle(0, 0, playfield->fieldSize.x, playfield->fieldSize.y, COLOR_FROM_HEX(0x225588));
    rlPushMatrix();
    rlTranslatef(-playfield->fieldSize.x / 2, -playfield->fieldSize.y / 2, 0);
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        Vector3 position = {bubble->position.x, bubble->position.y, 0};
        DrawModel(bubbleModel, position, bubble->radius + bubbleExtraRadius, bubbleTypeColors[bubble->bubbleType]);
        // DrawCircleLinesV(bubble->position, bubble->radius, COLOR_FROM_HEX(0xaaccff));
        // const char* bubbleLevel = TextFormat("%d:%.1f", bubble->bubbleLevel, bubble->bubbleMergeCooldown);
        // float width = MeasureText(bubbleLevel, 20);
        // DrawText(bubbleLevel, bubble->position.x - width / 2, bubble->position.y - 10, 20, COLOR_FROM_HEX(0xaaccff));
    }
    rlPopMatrix();
}

int main(void)
{
    // Initialization
    //--------------------------------------------------------------------------------------
    const int screenWidth = 800;
    const int screenHeight = 450;

    InitWindow(screenWidth, screenHeight, "raylib [core] example - basic window");

    SetTargetFPS(60);               // Set our game to run at 60 frames-per-second
    //--------------------------------------------------------------------------------------

    Camera3D camera = { 
        .position = { 0.0f, 0.0f, -10.0f },
        .target = { 0.0f, 0.0f, 0.0f },
        .up = { 0.0f, 1.0f, 0.0f },
        .projection = CAMERA_ORTHOGRAPHIC,
        .fovy = 350.0f,
    };
    Playfield playfield = {
        .fieldSize = {250, 300},
        .waterLevel = 170.0f,
    };

    bubbleModel = LoadModelFromMesh(GenMeshSphere(1.0f, 4, 24));
    // Main game loop
    while (!WindowShouldClose())    // Detect window close button or ESC key
    {
        float dt = GetFrameTime();
        // clamp dt to prevent large time steps, e.g. when browser tab is inactive
        if (dt > 0.2f) dt = 0.2f;
        gameTime.time += dt;
        gameTime.deltaTime = dt;


        if (IsMouseButtonReleased(MOUSE_LEFT_BUTTON))
        {
            PlayfieldTryAddBubble(&playfield, PlayfieldScreenToSpawnPosition(&playfield, camera, 
                GetMousePosition()));
        }
        // Draw
        //----------------------------------------------------------------------------------
        BeginDrawing();

        ClearBackground(COLOR_FROM_HEX(0x4488cc));

        BeginMode3D(camera);
        PlayfieldDraw(&playfield, camera);
        
        // fixed step update in post draw to allow debug drawing
        while (gameTime.fixedTime < gameTime.time)
        {
            gameTime.fixedTime += gameTime.fixedDeltaTime;
            PlayfieldFixedUpdate(&playfield);
        }

        EndMode3D();

        EndDrawing();
        //----------------------------------------------------------------------------------
    }

    // De-Initialization
    //--------------------------------------------------------------------------------------
    CloseWindow();        // Close window and OpenGL context
    //--------------------------------------------------------------------------------------

    return 0;
}